package org.nakedobjects.nos.client.web;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectApplicationException;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.context.Perspective;
import org.nakedobjects.nof.core.system.Monitor;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.Dump;
import org.nakedobjects.nos.client.web.component.Block;
import org.nakedobjects.nos.client.web.component.Component;
import org.nakedobjects.nos.client.web.component.ComponentFactory;
import org.nakedobjects.nos.client.web.component.DebugPane;
import org.nakedobjects.nos.client.web.component.ImageLookup;
import org.nakedobjects.nos.client.web.component.Page;
import org.nakedobjects.nos.client.web.object.CollectionView;
import org.nakedobjects.nos.client.web.object.FieldCollectionView;
import org.nakedobjects.nos.client.web.object.ObjectView;
import org.nakedobjects.nos.client.web.object.RemoveItemFromCollection;
import org.nakedobjects.nos.client.web.object.Save;
import org.nakedobjects.nos.client.web.object.ServiceView;
import org.nakedobjects.nos.client.web.request.Action;
import org.nakedobjects.nos.client.web.request.ActionException;
import org.nakedobjects.nos.client.web.request.ChangeContext;
import org.nakedobjects.nos.client.web.request.Context;
import org.nakedobjects.nos.client.web.request.Crumb;
import org.nakedobjects.nos.client.web.request.ObjectLookupException;
import org.nakedobjects.nos.client.web.request.Request;
import org.nakedobjects.nos.client.web.request.TaskLookupException;
import org.nakedobjects.nos.client.web.servlet.SystemAccess;
import org.nakedobjects.nos.client.web.session.LogOut;
import org.nakedobjects.nos.client.web.session.Welcome;
import org.nakedobjects.nos.client.web.task.AddItemToCollection;
import org.nakedobjects.nos.client.web.task.EditObject;
import org.nakedobjects.nos.client.web.task.InvokeMethod;
import org.nakedobjects.nos.client.web.task.TaskStep;


public class WebController {
    private static final String ERROR_REASON = "This error occurs when you go back to a page " +
                        		"using the browsers back button.  To avoid this error in the future please avoid using the back button";

    private class DebugView implements Action {
        public void execute(final Request request, final Context context, final Page page) {
            page.setTitle("Debug");

            DebugPane debugPane = context.getComponentFactory().createDebugPane();
            page.setDebug(debugPane);
            
            DebugString debug = new DebugString();

            NakedObject perspective = NakedObjectsContext.getPerspective();
            debug.appendTitle("Perspective");
            if (perspective != null && perspective.getObject() != null) {
                debug.appendln(perspective.titleString());
                Perspective p = ((Perspective) perspective.getObject());
                debug.appendln("name", p.getName());
                debug.appendln("user", p.getUserName());
                debug.appendln("services", p.getServices().toString());
            } else {
                debug.appendln("none");
            }

            debug.appendTitle("Actions");
            Iterator e = actions.entrySet().iterator();
            debug.indent();
            while (e.hasNext()) {
                Map.Entry element = (Map.Entry) e.next();
                debug.appendln(element.getKey() + " -> " + element.getValue());
            }
            debug.unindent();

            context.debug(debug);

            ImageLookup.debug(debug);
            
            debugPane.appendln("<pre>" + debug.toString() + "</pre>");
        }

        public String name() {
            return "debug";
        }
    }
    
    private class DebugSpecification implements Action {
        public void execute(Request request, Context context, Page page) {
            DebugPane debugPane = context.getComponentFactory().createDebugPane();
            page.setDebug(debugPane);

            debugPane.addSection("Specification");
            NakedObject object = context.getMappedObject(request.getObjectId());
            debugPane.appendln(Dump.specification(object));
        }

        public String name() {
            return "spec";
        }
    }
    
    private class DebugObject implements Action {
        public void execute(Request request, Context context, Page page) {
            DebugPane debugPane = context.getComponentFactory().createDebugPane();
            page.setDebug(debugPane);

            debugPane.addSection("Adapter");
            NakedObject object = context.getMappedObject(request.getObjectId());
            debugPane.appendln(Dump.adapter(object));
            debugPane.addSection("Graph");
            debugPane.appendln(Dump.graph(object));
        }

        public String name() {
            return "dump";
        }
    }

    private Map actions = new HashMap();
    private boolean isDebug;
    private final Logger LOG = Logger.getLogger(WebController.class);
    private final Logger ACCESS_LOG = Logger.getLogger("access_log");

    public boolean actionExists(final Request req) {
        return actions.containsKey(req.getRequestType());
    }

    protected void addAction(final Action action) {
        actions.put(action.name(), action);
    }

    private void addCrumbs(final Context context, final Page page) {
        Crumb[] crumbs = context.getCrumbs();
        String[] names = new String[crumbs.length];
        boolean[] isLinked = context.isLinked();

        for (int i = 0; i < crumbs.length; i++) {
            names[i] = crumbs[i].title();
        }

        ComponentFactory factory = context.getComponentFactory();
        Component breadCrumbs = factory.createBreadCrumbs(names, isLinked);

        page.setCrumbs(breadCrumbs);
    }

    public void addDebug(final Page page, Request req) {
        page.addDebug("<a href=\"debug.app\">Debug</a>");
        String id = req.getObjectId();
        if(id != null) {
            page.addDebug("<a href=\"dump.app?id=" + id + "\">Object</a>");
            page.addDebug("<a href=\"spec.app?id=" + id + "\">Spec</a>");
        }
        page.addDebug("<a href=\"about.app\">About</a>");
    }

    public Page generatePage(final Context context, final Request request) {
        context.restoreAllObjectsToLoader();
        Page page = context.getComponentFactory().createPage();
        pageHeader(context, page);
        Block navigation = page.getNavigation();

        Block optionBar = context.getComponentFactory().createBlock("options", null);
        optionBar.add(context.getComponentFactory().createHeading("Options"));
        
        Block block = context.getComponentFactory().createBlock("item", null);
        Component option = context.getComponentFactory().createLink("logout", "Log Out", "End the current session");
        block.add(option);
        optionBar.add(block);

        block = context.getComponentFactory().createBlock("item", null);
        option = context.getComponentFactory().createLink("about", "About", "Details about this application");
        block.add(option);
        optionBar.add(block);

        if (SystemAccess.showExploration()) {
            block = context.getComponentFactory().createBlock("item", null);
            option = context.getComponentFactory().createLink("swapuser", "Swap User", "Swap the exploration user");
            block.add(option);
            optionBar.add(block);
        }
        
        navigation.add(optionBar);

        listServices(context, navigation);
        listHistory(context, navigation);
        Monitor.addEvent("Web", "Request " + request);
        runAction(context, request, page);
        addCrumbs(context, page);
        
        // The web viewer has no views of other objects, so changes can be ignored
        NakedObjectsContext.getUpdateNotifer().allChangedObjects();
        // TODO deal with disposed objects
        
        return page;
    }

    public void init() {
        addAction(new About());
        addAction(new SwapUser());
        addAction(new SetUser());
        addAction(new DebugView());
        addAction(new DebugSpecification());
        addAction(new DebugObject());
        addAction(new Welcome());
        addAction(new ObjectView());
        addAction(new CollectionView());
        addAction(new FieldCollectionView());
        addAction(new InvokeMethod());
        addAction(new TaskStep());
        addAction(new EditObject());
        addAction(new Save());
        addAction(new ServiceView());
        addAction(new LogOut());
        addAction(new RemoveItemFromCollection());
        addAction(new AddItemToCollection());
        addAction(new ChangeContext());

        Logger.getLogger(this.getClass()).info("init");
    }

    public boolean isDebug() {
        return isDebug;
    }
    
    private void listHistory(final Context context, final Block navigation) {
        context.listHistory(context, navigation);
    }

    private void listServices(final Context context, final Block navigationBar) {
        NakedObject perspective = NakedObjectsContext.getPerspective();
        Block taskBar = context.getComponentFactory().createBlock("services", null);
        taskBar.add(context.getComponentFactory().createHeading("Services"));
        NakedObjectField fld = perspective.getSpecification().getField("services");
        NakedObjectsContext.getObjectPersistor().resolveField(perspective, fld);
        NakedCollection classList = ((NakedCollection) fld.get(perspective));
        Object[] services = new Object[classList.size()];
        int i = 0;
        for (Enumeration e = classList.elements(); e.hasMoreElements();) {
            NakedObject adapter = (NakedObject) e.nextElement();
            services[i] = adapter.getObject();
            String id = context.mapObject(adapter);
            Component service = context.getComponentFactory().createService(id, adapter.titleString(), adapter.getIconName());
            taskBar.add(service);
            i++;
        }
        navigationBar.add(taskBar);
    }

    private void pageHeader(final Context context, final Page page) {
        page.getPageHeader().add(context.getComponentFactory().createInlineBlock("none", "", null));
    }

    private void runAction(final Context context, final Request request, final Page page) {
        try {
            ACCESS_LOG.info("request " + request.toString());
            Request r = request;
            DebugString debug = new DebugString();
            debug.startSection("Request");
            debug.appendln("http", request.toString());
            debug.endSection();
            do {
                Action action = (Action) actions.get(r.getRequestType());
                try {
                    action.execute(r, context, page);
                } catch (ObjectLookupException e) {
                    String error = "The object/service you selected has timed out.  Please navigate to the object via the history bar.";
                    displayError(context, page, error);
                } catch (TaskLookupException e) {
                    String error = "The task you went back to has already been completed or cancelled.  Please start the task again.";
                    displayError(context, page, error);
                }
                r = r.getForward();
                if (r != null) {
                    LOG.debug("forward to " + r);
                }
            } while (r != null);
            if (LOG.isDebugEnabled()) {
                context.debug(debug);
                debug.appendln();
                NakedObjectsContext.getDebugContext(NakedObjectsContext.getExecutionContextId()).debugAll(debug);
                LOG.debug(debug.toString());
            }
        } catch (ActionException e) {
            page.setTitle("Error");
            page.getViewPane().setTitle("Error", "Action Exception");
            LOG.error("ActionException, executing action " + request.getRequestType(), e);
            page.getViewPane().add(context.getComponentFactory().createErrorMessage(e, isDebug));
        } catch (NakedObjectApplicationException e) {
            page.setTitle("Error");
            page.getViewPane().setTitle("Error", "Application Exception");
            LOG.error("ApplicationException, executing action " + request.getRequestType(), e);
            page.getViewPane().add(context.getComponentFactory().createErrorMessage(e, isDebug));
        } catch (NakedObjectRuntimeException e) {
            page.setTitle("Error");
            page.getViewPane().setTitle("Error", "System Exception");
            LOG.error("NakedObjectRuntimeException, executing action " + request.getRequestType(), e);
            page.getViewPane().add(context.getComponentFactory().createErrorMessage(e, true));
        } catch (RuntimeException e) {
            page.setTitle("Error");
            page.getViewPane().setTitle("Error", "System Exception");
            LOG.error("RuntimeException, executing action " + request.getRequestType(), e);
            page.getViewPane().add(context.getComponentFactory().createErrorMessage(e, true));
        }
    }

    private void displayError(final Context context, final Page page, String errorMessage) {
        page.setTitle("Error");
        page.getViewPane().setTitle("Error", "");
        
        Block block1 = context.getComponentFactory().createBlock("error", "");
        block1.add(context.getComponentFactory().createInlineBlock("", errorMessage, ""));
        page.getViewPane().add(block1);
        
        Block block2 = context.getComponentFactory().createBlock("text", "");
        block2.add(context.getComponentFactory().createInlineBlock("", ERROR_REASON, ""));
        page.getViewPane().add(block2);
    }
    
    public void setDebug(boolean on) {
        this.isDebug = on;
    }
}

// Copyright (c) Naked Objects Group Ltd.
