package org.nakedobjects.nos.client.web.component;

import java.io.File;
import java.io.FilenameFilter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.util.DebugString;


/**
 * ImageLookup provides an efficient way of finding the most suitable image to use. It ensures that an image
 * is always available, providing a default image if needed. All requests are cached to improve performance.
 */
// TODO allow for multiple extension types
public class ImageLookup {
    private static final String UNKNOWN_IMAGE = "Default";
    private static final String[] EXTENSIONS = { "png", "gif", "jpg", "jpeg" };
    private static final Map images = new HashMap();
    private static File imageDirectory;
    private static String unknownImageFile;

    public static void setImageDirectory(String imageDirectory) {
        ImageLookup.imageDirectory = new File(imageDirectory);
        if(!ImageLookup.imageDirectory.exists()) {
            throw new NakedObjectRuntimeException("No image directory found: " + imageDirectory);
        }
        unknownImageFile = imageFile(UNKNOWN_IMAGE);
    }
    
    public static void debug(DebugString debug) {
        debug.appendTitle("Image Lookup");
        debug.indent();
        Iterator keys = images.keySet().iterator();
        while (keys.hasNext()) {
            Object key = keys.next();
            Object value = images.get(key);
            debug.appendln(key + " -> " + value);
        }
        debug.unindent();
    }

    private static String imageFile(final String imageName) {
        String[] files = imageDirectory.list(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                int dot = name.lastIndexOf('.');
                if (dot > 0) {
                    for (int i = 0; i < EXTENSIONS.length; i++) {
                        if (name.substring(0, dot).equalsIgnoreCase(imageName) && name.substring(dot + 1).equalsIgnoreCase(EXTENSIONS[i])) {
                            return true;
                        }
                    }
                }
                return false;
            }
        });
        
        return files.length == 0 ? null : files[0];
    }

    private static String findImage(NakedObjectSpecification specification) {
        String name = specification.getShortName();
        String fileName = imageFile(name);
        if (fileName != null) {
            images.put(name, fileName);
            return fileName;
        } else {
            NakedObjectSpecification superclass = specification.superclass();
            if (superclass != null) {
                return findImage(superclass);
            } else {
                return unknownImageFile;
            }
        }
    }

    /**
     * For an object, the icon name from the object is return if it is not null, otherwise the specification
     * is used to look up a suitable image name.
     * 
     * @see NakedObject#getIconName()
     * @see #image(NakedObjectSpecification)
     */
    public static String image(NakedObject object) {
        String iconName = object.getIconName();
        if (iconName != null) {
            return image(iconName);
        } else {
            return image(object.getSpecification());
        }
    }

    public static String image(NakedObjectSpecification specification) {
        String name = specification.getShortName();
        String imageName = (String) images.get(name);
        if (imageName != null) {
            return (String) imageName;
        } else {
            return findImage(specification);
        }
    }

    public static String image(String name) {
        String imageName = (String) images.get(name);
        if (imageName != null) {
            return (String) imageName;
        } else {
            String fileName = imageFile(name);
            return fileName == null ? unknownImageFile : fileName;
        }
    }

}

// Copyright (c) Naked Objects Group Ltd.
