package org.nakedobjects.nos.client.web.html;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;

import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.value.BooleanValue;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.NakedObjectAction.Type;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.NakedObjectConfiguration;
import org.nakedobjects.nos.client.web.component.Block;
import org.nakedobjects.nos.client.web.component.Component;
import org.nakedobjects.nos.client.web.component.ComponentFactory;
import org.nakedobjects.nos.client.web.component.DebugPane;
import org.nakedobjects.nos.client.web.component.Form;
import org.nakedobjects.nos.client.web.component.Page;
import org.nakedobjects.nos.client.web.component.Table;
import org.nakedobjects.nos.client.web.request.WebViewerException;


public class HtmlComponentFactory implements ComponentFactory {
    private static final String PROPERTY_BASE = Configuration.ROOT + "viewer.web.";
    private String footer;
    private String header;
    private String styleSheet;

    public HtmlComponentFactory() {
        NakedObjectConfiguration configuration = NakedObjectsContext.getConfiguration();
        styleSheet = configuration.getString(PROPERTY_BASE + "style-sheet");
        String file = configuration.getString(PROPERTY_BASE + "header-file");
        header = file == null ? configuration.getString(PROPERTY_BASE + "header") : loadFile(file);
        file = configuration.getString(PROPERTY_BASE + "footer-file");
        footer = file == null ? configuration.getString(PROPERTY_BASE + "footer") : loadFile(file);
    }

    public Block createBlock(final String style, final String description) {
        return new Div(style, description);
    }

    public Component createBreadCrumbs(final String[] names, final boolean[] isLinked) {
        return new BreadCrumbs(names, isLinked);
    }

    public Component createCollectionIcon(final NakedObjectField field, final NakedCollection collection, final String id) {
        return new CollectionLink(field, collection, field.getDescription(), id);
    }

    public DebugPane createDebugPane() {
        return new HtmlDebug();
    }

    public Component createEditOption(final String id) {
        return new Action("edit", "Edit Object", "Edit the current object", id, null, null);
    }

    public Component createRemoveOption(final String id, final String elementId, final String fieldName) {
        return new Action("remove", "Remove", "Remove item from collection", id, elementId, fieldName);
    }

    public Component createAddOption(final String id, final String fieldName) {
        return new Action("add", "Add Item", "Add item to collection", id, null, fieldName);
    }

    public Component createErrorMessage(final Exception e, final boolean isDebug) {
        return new ErrorMessage(e, isDebug);
    }

    public Form createForm(final String id, final String actionName, final int step, final int noOfPages, final boolean isEditing) {
        return new HtmlForm(id, actionName, step, noOfPages, isEditing);
    }

    public Component createHeading(final String name) {
        return new Heading(name, 4);
    }

    public Component createInlineBlock(final String style, final String text, final String description) {
        return new Span(style, text, description);
    }

    public Component createSubmenu(final String menuName, final Component[] items) {
        return new Submenu(menuName, items);
    }

    public Component createLink(String link, final String name, final String description) {
        return new Link(link, name, description);
    }
    
    public Component createMenuItem(
            final String actionId,
            final String name,
            final String description,
            final String status,
            final Type type,
            final boolean hasParameters,
            final String targetObjectId) {
        return new MenuItem(actionId, name, description, status, type, hasParameters, targetObjectId);
    }

    public Component createCollectionIcon(final NakedCollection collection, final String collectionId) {
        return new CollectionIcon(collection, collection.getSpecification().getDescription(), collectionId);
    }

    public Component createObjectIcon(final NakedObject object, final String objectId, final String style) {
        return new ObjectIcon(object, object.getSpecification().getDescription(), objectId, style);
    }

    public Component createObjectIcon(NakedObjectField field, NakedObject object, String objectId, String style) {
        return new ObjectIcon(object, field.getDescription(), objectId, style);
    }

    public Page createPage() {
        return new DynamicHtmlPage(styleSheet, header, footer);
    }
    
    public LogonFormPage createLogonPage(final String user, String password) {
        return new LogonFormPage(styleSheet, header, footer, user, password);
    }

    public Component createService(final String objectId, final String title, final String iconName) {
        return new Service(objectId, title, iconName);
    }

    public Table createTable(final int noColumns, final boolean withSelectorColumn) {
        return new HtmlTable(noColumns, withSelectorColumn);
    }

    public Component createUserSwap(final String name) {
        return new Component() {
            public void write(PrintWriter writer) {
                writer.print("<a class=\"user\" href=\"setuser.app?name=");
                writer.print(name);
                writer.print("\" title=\"Change user to " + name);
                writer.print("\">");
                writer.print(name);
                writer.println("</a>");
            }
        };
        
    }
  
    public Component createValueField(final NakedObjectField field, final NakedValue value, final boolean isWrapped, boolean wrap, final boolean isEditable) {
        if (value instanceof BooleanValue) {
            // TODO extract out this mechanism so it can be generalised
            return new Component() {
                public void write(PrintWriter writer) {
                    writer.print("<input class=\"value\" type=\"checkbox\"");
                    if (((BooleanValue) value).isSet()) {
                        writer.print(" checked");
                    }
                    if (!isEditable) {
                        writer.print(" disabled");
                    }
                    writer.println("/>");
                }
            };
        } else {
            if (isWrapped) {
                return createInlineBlock("value", "<pre>" + value.titleString() + "</pre>", null);
            } else {
                return createInlineBlock("value", value.titleString(), null);
            }
        }
    }

    private String loadFile(final String file) {
        StringBuffer content = new StringBuffer();
        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
            String line;
            while ((line = reader.readLine()) != null) {
                content.append(line);
                content.append("\n");
            }
        } catch (FileNotFoundException e) {
            throw new WebViewerException("Failed to find file " + file);
        } catch (IOException e) {
            throw new WebViewerException("Failed to load file " + file, e);
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (IOException ignore) {}
            }
        }
        return content.toString();
    }
}

// Copyright (c) Naked Objects Group Ltd.
