package org.nakedobjects.nos.client.web.request;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.noa.reflect.NakedObjectAction;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nos.client.web.component.Block;
import org.nakedobjects.nos.client.web.component.ComponentFactory;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Map.Entry;

import org.apache.log4j.Logger;


public class Context {
    private static final Logger LOG = Logger.getLogger(Context.class);
    private final Map actionMap = new HashMap();
    private final ComponentFactory componentFactory;
    private final ObjectHistory history = new ObjectHistory();
    private boolean isValid;
    private int max;
    private final Map collectionMap = new HashMap();
    private final Map objectMap = new HashMap();
    private final Map serviceMap = new HashMap();
    private final Stack crumbs = new Stack();
    private Session session;
    private String[] messages;
    private String[] warnings;

    
    public Context(final ComponentFactory factory) {
        componentFactory = factory;
        isValid = true;
        clearMessagesAndWarnings();
        LOG.debug(this + " created with " + factory);
    }

    public void setObjectCrumb(final NakedObject object) {
        while(crumbs.size() > 0 && !(crumbs.lastElement() instanceof TaskCrumb)) {
            crumbs.pop();
        }
        String id = mapObject(object);
        crumbs.push(new ObjectCrumb(id, object));
    }

    public void addCollectionFieldCrumb(final String collectionFieldName) {
        crumbs.push(new ObjectFieldCrumb(collectionFieldName));
    }

    public void addCollectionCrumb(final String id) {
        while(crumbs.size() > 0 && !(crumbs.lastElement() instanceof TaskCrumb)) {
            crumbs.pop();
        }
        crumbs.push(new CollectionCrumb(id, getMappedCollection(id)));
    }

    public void addTaskCrumb(final Task task) {
        while(crumbs.size() > 1 && !(crumbs.lastElement() instanceof ObjectCrumb)) {
            crumbs.pop();
        }
        Assert.assertNotNull(task);
        Assert.assertTrue(!isTask());
        task.init(this);
        crumbs.push(new TaskCrumb(task));
    }

    public void debug(final DebugString debug) {
        debug.startSection("Web Session Context");
        debug.appendAsHexln("hash", hashCode());
        debug.appendln("session", session);
        debug.appendln("is valid", isValid);
        debug.appendln("next id", max);
        debug.appendln("factory", componentFactory);
        debug.appendln("is task", isTask());
        
        debug.appendln("crumbs (" + crumbs.size() + ")");
        debug.indent();
        for (int i = 0; i < crumbs.size(); i++) {
            Crumb crumb = (Crumb) crumbs.get(i);
            debug.appendln(i  + 1 + ". " + crumb);
            debug.indent();
            crumb.debug(debug);
            debug.unindent();
        }
        debug.unindent();
                
        debug.startSection("Objects");
        Iterator ids = objectMap.keySet().iterator();
        while (ids.hasNext()) {
            String id = (String) ids.next();
            ObjectMapping object = (ObjectMapping) objectMap.get(id);
            debug.appendln(id + " -> " + object.getOid());
            debug.indent();
            object.debug(debug);
            debug.unindent();
        }
        debug.endSection();
        
        debug.startSection("Collections");
        ids = collectionMap.keySet().iterator();
        while (ids.hasNext()) {
            String id = (String) ids.next();
            CollectionMapping coll = (CollectionMapping) collectionMap.get(id);
            debug.appendln(id + " -> collection of " + coll.getElementSpecification().getPluralName());
            coll.debug(debug);
        }
        debug.endSection();
        
        debug.startSection("Actions");
        debugMap(debug, actionMap);
        debug.endSection();

        debug.startSection("History");
        history.debug(debug);
        debug.endSection();

        debug.endSection();
    }

    private void debugMap(final DebugString debug, final Map map) {
        Iterator names = map.keySet().iterator();
        while (names.hasNext()) {
            String name = (String) names.next();
            debug.appendln(name + " -> " + map.get(name));
        }
    }

    public NakedObjectAction getMappedAction(final String id) {
        return (NakedObjectAction) getMappedInstance(actionMap, id);
    }

    public ComponentFactory getComponentFactory() {
        return componentFactory;
    }

    /**
     * Returns an array of instances of the specified type that are currently known in the current context,
     * ie have been recently seen by the user.
     */
    public NakedObject[] getKnownInstances(final NakedObjectSpecification type) {
        Iterator ids = objectMap.keySet().iterator();
        List instances = new ArrayList();
        while (ids.hasNext()) {
            String id = (String) ids.next();
            NakedObject object = (NakedObject) getMappedObject(id);
            if (object.getSpecification().isOfType(type)) {
                instances.add(object);
            }
        }

        NakedObject[] array = new NakedObject[instances.size()];
        instances.toArray(array);
        return array;
    }

    private String addToMap(final Map map, final Object object) {
        Assert.assertNotNull(object);
        if (map.containsValue(object)) {
           return findExistingId(map, object);
        } else {
            return mapNewObject(map, object);
        }
    }

    private String mapNewObject(final Map map, final Object object) {
        max++;
        String id = "" + max;
        map.put(id, object);
        
        String mapName = map == objectMap ? "object" : (map == collectionMap ? "collection" : "action"); 
        LOG.debug("add " + object + " to " + mapName + " as #" + id);
        
        return id;
    }

    private String findExistingId(final Map map, final Object object) {
        Set set = map.keySet();
        Iterator iterator = set.iterator();
        while (iterator.hasNext()) {
            String id = (String) iterator.next();
            if (object.equals(map.get(id))) {
                return id;
            }
        }
        throw new NakedObjectRuntimeException();
    }

    private Object getMappedInstance(final Map map, final String id) {
        Object object = map.get(id);
        if (object == null) {
            String mapName = map == objectMap ? "object" : (map == collectionMap ? "collection" : "action"); 
            throw new ObjectLookupException("No object in " + mapName + " map with id " + id);
        }
        return object;
    }

    public NakedCollection getMappedCollection(final String id) {
        CollectionMapping map = (CollectionMapping) getMappedInstance(collectionMap, id);
        return map.getCollection(this);
    }

    public NakedObject getMappedObject(final String id) {
        ObjectMapping mappedObject = (ObjectMapping) getMappedInstance(objectMap, id);
        NakedObject object = mappedObject.getObject();
        try {
            mappedObject.checkVersion(object);
        } catch (ConcurrencyException e) {
            LOG.info("concurrency conflict: " + e.getMessage());
            messages = new String[] { e.getMessage(), "Reloaded object " + object.titleString()};
            updateVersion(object);
        }
        return object;
    }

    public Task getTask(final String taskId) {
        Task task = null;
        for (int i = crumbs.size() - 1; i >= 0; i--) {
            Object crumb = crumbs.get(i);
            if (crumb instanceof TaskCrumb) {
                TaskCrumb taskCrumb = (TaskCrumb) crumb;
                String id = taskCrumb.getTask().getId();
                if (taskId.equals(id)) {
                    task = taskCrumb.getTask();
                    break;
                }
            }
        }
        return task;
    }
    
    public Request cancelTask(final Task task) {
        if (task != null) {
            endTask(task);
        }
        
        // REVIEW does this take us back to the right object?
        Crumb crumb = (Crumb) crumbs.get(crumbs.size() - 1);
        return crumb.changeContext();
    }

    public void invalidate() {
        isValid = false;
    }

    public boolean isValid() {
        return isValid;
    }
    
    public boolean isLoggedIn() {
        return session != null;
    }

    public String mapAction(final NakedObjectAction action) {
        return addToMap(actionMap, action);
    }

    public String mapObject(final NakedObject object) {
        ObjectMapping mapping = null;
        if (object.getResolveState().isTransient()) {
            mapping = new TransientObjectMapping(object);
        } else {
            mapping = new PersistentObjectMapping(object);
        }
        return addToMap(objectMap, mapping);
    }

    public String mapCollection(final NakedCollection collection) {
        CollectionMapping map = new CollectionMapping(this, collection);
        return addToMap(collectionMap, map);
    }
    
    public void endTask(final Task task) {
        for (int i = crumbs.size() - 1; i >= 0; i--) {
            Object crumb = crumbs.get(i);
            if (crumb instanceof TaskCrumb) {
                TaskCrumb taskCrumb = (TaskCrumb) crumb;
                if (taskCrumb.getTask() == task) {
                    crumbs.remove(taskCrumb);
                    return;
                }
            }
        }
        throw new NakedObjectRuntimeException("No crumb found for " + task);
    }

    public Crumb[] getCrumbs() {
        int size = crumbs.size();
        Crumb[] taskList = new Crumb[size];
        for (int i = 0; i < crumbs.size(); i++) {
            taskList[i] = (Crumb) crumbs.get(i);
        }
        return taskList;
    }
    
    public boolean[] isLinked() {
        int size = crumbs.size();
        boolean[] isLinked = new boolean[size];
        for (int i = size - 1; i >=0 ; i--) {
            boolean isTask = crumbs.elementAt(i) instanceof TaskCrumb;
            isLinked[i] = i != size - 1;
            if(isTask) {
                break;
            }
        }
        return isLinked;
    }

    public String[] getMessages() {
        return messages;
    }
    
    public String[] getWarnings() {
        return warnings;
    }    
    
    public void setMessagesAndWarnings(String[] messages, String[] warnings) {
        this.messages = messages;
        this.warnings = warnings;
    }
    
    public void clearMessagesAndWarnings() {
      messages = warnings = new String[0];
    }
    
    private boolean isTask() {
        int index = crumbs.size() - 1;
        return index >= 0 && crumbs.get(index) instanceof TaskCrumb;
    }

    public Request changeContext(final int id) {
        while(crumbs.size() - 1 > id) {
            crumbs.pop();
        }
        Crumb c = (Crumb) crumbs.lastElement();
        return c.changeContext();
    }

    public void setSession(Session currentSession) {
        this.session = currentSession;
    }
    
    public Session getSession() {
        return session;
    }

    public void purge() {
        actionMap.clear();
        clearMessagesAndWarnings();
        //purgeCollections();
        //purgeObjects();
        
        Map newCollectionMap = new HashMap();
        Map newObjectMap = new HashMap();
        
        Enumeration e = history.elements();
        while (e.hasMoreElements()) {
            HistoryEntry entry = (HistoryEntry) e.nextElement();
            if (entry.type == HistoryEntry.OBJECT) {
                Object item = objectMap.get(entry.id);
                newObjectMap.put(entry.id, item);    
                LOG.debug("copied object map " + entry.id + " for " + item);
                ((ObjectMapping) item).updateVersion();
            } else if (entry.type == HistoryEntry.COLLECTION) {
                CollectionMapping coll = (CollectionMapping) collectionMap.get(entry.id);
                newCollectionMap.put(entry.id, coll);
                LOG.debug("copied collection map for " + coll);
                Enumeration e1 = coll.elements();
                while (e1.hasMoreElements()) {
                    String id1 = (String) e1.nextElement();
                    Object item = objectMap.get(id1);
                    newObjectMap.put(id1, item);  
                    LOG.debug("copied object map " + id1 + " for " + item);
                    ((ObjectMapping) item).updateVersion();
                }
            }
        }
        
        collectionMap.clear();
        collectionMap.putAll(newCollectionMap);
        objectMap.clear();
        objectMap.putAll(newObjectMap);
        objectMap.putAll(serviceMap);
    }
    
    public void restoreAllObjectsToLoader() {
        Set oidSet = objectMap.entrySet();
        for(Iterator it = oidSet.iterator(); it.hasNext();) {
            ObjectMapping mapping = (ObjectMapping) ((Entry) it.next()).getValue();
            mapping.restoreToLoader(NakedObjectsContext.getObjectLoader());
        }
    }
    
    public void listHistory(Context context, Block navigation) {
        history.listObjects(context, navigation);
    }

    public void addObjectToHistory(String idString) {
        history.addObject(idString);
    }

    public void addCollectionToHistory(String idString) {
        history.addCollection(idString);
    }

    public void init() {
        NakedObject perspective = NakedObjectsContext.getPerspective();
        NakedObjectField fld = perspective.getSpecification().getField("services");
        NakedCollection classList = ((NakedCollection) fld.get(perspective));
        for (Enumeration e = classList.elements(); e.hasMoreElements();) {
            NakedObject adapter = (NakedObject) e.nextElement();
            mapObject(adapter);
        }
        serviceMap.putAll(objectMap);
    }

    public void updateVersion(NakedObject object) {
        if (object.getResolveState().isTransient()) {
            return;
        }
       
        // TODO refactor this for clarity: removes existing mapping and replaces it with a new one as it contains the new version
        String id = mapObject(object);
        if (id != null) {
            ObjectMapping mapping = new PersistentObjectMapping(object);
            objectMap.put(id, mapping);
        }
    }
}

// Copyright (c) Naked Objects Group Ltd.
