package org.nakedobjects.nos.client.web.request;

import org.nakedobjects.noa.adapter.InvalidEntryException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.TextEntryParseException;
import org.nakedobjects.noa.adapter.value.PasswordValue;
import org.nakedobjects.noa.spec.Features;
import org.nakedobjects.noa.spec.HasType;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nos.client.web.component.Page;


/**
 * Represents a task that the user is working through. Is used for both editing objects and setting up
 * parameters for an action method.
 */
public abstract class Task {
    private static int nextID = 1;
    private int[] boundaries;
    private final String description;
    protected final String[] errors;
    protected String error;
    private final String[] entryText;
    protected final Naked[] initialState;
    private final String name;
    protected final String[] names;
    protected final String[] descriptions;
    protected final boolean[] optional;
    protected final boolean[] readOnly;
    protected final int numberOfEntries;
    private int step;
    private final String targetId;
    protected final NakedObjectSpecification[] fieldSpecifications;
    protected final int[] noLines;
    protected final boolean[] wraps;
    protected final int[] maxLength;
    protected final int[] typicalLength;
    protected final int id = nextID++;

    public Task(final Context context, final String name, final String description, final NakedObject target, final int noFields) {
        this.name = name;
        this.description = description;
        targetId = context.mapObject(target);

        initialState = new Naked[noFields];
        names = new String[noFields];
        descriptions = new String[noFields];
        optional = new boolean[noFields];
        readOnly = new boolean[noFields];
        fieldSpecifications = new NakedObjectSpecification[noFields];

        numberOfEntries = noFields;
        entryText = new String[noFields];
        errors = new String[noFields];

        noLines = new int[noFields];
        wraps = new boolean[noFields];
        maxLength = new int[noFields];
        typicalLength = new int[noFields];
    }

    void init(Context context) {
        for (int i = 0; i < entryText.length; i++) {
            Naked obj = initialState[i];
            if (obj == null) {
                entryText[i] = "";
            } else if (obj.getSpecification().getType() == HasType.VALUE) {
                if (PasswordValue.class.isAssignableFrom(obj.getClass())) {
                    entryText[i] = ((PasswordValue) obj).getEditText();
                } else {
                    entryText[i] = obj.titleString();
                }
            } else if (obj.getSpecification().getType() == HasType.OBJECT) {
                if (readOnly[i]) {
                    entryText[i] = ((NakedObject) obj).titleString();
                } else {
                    entryText[i] = context.mapObject((NakedObject) obj);
                }
            } else if (obj.getSpecification().getType() == HasType.COLLECTION) {
                entryText[i] = ((NakedCollection) obj).titleString();
            }
        }

        divyUpWork();
    }

    public abstract Naked completeTask(Context context, Page page);

    private void copyForThisStep(Object[] source, Object[] destination) {
        for (int i = 0; i < noOfEntriesInThisStep(); i++) {
            destination[i] = source[firstEntryInThisStep() + i];
        }
    }

    private void copyForThisStep(boolean[] source, boolean[] destination) {
        for (int i = 0; i < noOfEntriesInThisStep(); i++) {
            destination[i] = source[firstEntryInThisStep() + i];
        }
    }

    private void copyForThisStep(int[] source, int[] destination) {
        for (int i = 0; i < noOfEntriesInThisStep(); i++) {
            destination[i] = source[firstEntryInThisStep() + i];
        }
    }

    public void checkInstances(Context context, NakedObject[] objects) {}

    public void debug(DebugString debug) {
        debug.indent();
        debug.appendln("name", name);
        debug.appendln("number of steps ", numberOfSteps());
        debug.appendln("current step", step);
        debug.appendln("target", targetId);
        debug.appendln("steps (" + (boundaries.length - 1) + ")");
        debug.indent();
        for (int i = 0; i < boundaries.length - 1; i++) {
            debug.appendln("    " + (i + 1) + ". " + boundaries[i] + " - " + (boundaries[i + 1] - 1));
        }
        debug.unindent();
        debug.appendln("fields (" + names.length + ")");
        debug.indent();
        for (int i = 0; i < names.length; i++) {
            String status = (readOnly[i] ? "R" : "-") + (optional[i] ? "O" : "M") + (errors[i] == null ? "-" : "E");
            debug.appendln("    " + i + "  " + names[i] + " (" + status + "):  " + fieldSpecifications[i].getFullName() + " -> "
                    + entryText[i]);
        }
        debug.unindent();
        debug.unindent();
    }

    private void divyUpWork() {
        if (numberOfEntries == 0) {
            boundaries = new int[2];
        } else {
            int[] b = new int[numberOfEntries + 2];
            int count = 0;
            b[count++] = 0;

            NakedObjectSpecification type = fieldSpecifications[0];
            boolean direct = simpleField(type, 0);

            for (int i = 1; i < numberOfEntries; i++) {
                type = fieldSpecifications[i];
                if (true || direct && (simpleField(type, i))) {
                    continue;
                }
                b[count++] = i;
                direct = simpleField(type, i);
            }
            b[count++] = numberOfEntries;
            boundaries = new int[count];
            System.arraycopy(b, 0, boundaries, 0, count);
        }
    }

    protected boolean simpleField(final NakedObjectSpecification type, final int i) {
        return readOnly[i] || type.getType() == HasType.VALUE
                || (type.getType() == HasType.OBJECT && Features.isBoundedSet(type));
    }

    private int firstEntryInThisStep() {
        return boundaries[step];
    }

    public String getDescription() {
        return description;
    }

    public String getError() {
        return error;
    }

    /**
     * Returns an array of errors, one for each element in the task.
     */
    public String[] getErrors() {
        String[] array = new String[noOfEntriesInThisStep()];
        copyForThisStep(errors, array);
        return array;
    }

    public String[] getFieldDescriptions() {
        String[] array = new String[noOfEntriesInThisStep()];
        copyForThisStep(descriptions, array);
        return array;
    }

    public String[] getEntryText() {
        String[] array = new String[noOfEntriesInThisStep()];
        copyForThisStep(entryText, array);
        return array;
    }

    public String getName() {
        return name;
    }

    public String[] getNames() {
        String[] array = new String[noOfEntriesInThisStep()];
        copyForThisStep(names, array);
        return array;
    }

    public Naked[][] getOptions(final Context context) {
        return getOptions(context, firstEntryInThisStep(), noOfEntriesInThisStep());
    }

    protected Naked[][] getOptions(final Context context, final int from, final int len) {
        return new Naked[len][];
    }

    protected Naked[] getEntries(final Context context) {
        Naked[] entries = new Naked[entryText.length];
        for (int i = 0; i < entries.length; i++) {
            if (entryText == null || readOnly[i]) {
                continue;
            }
            NakedObjectSpecification spec = fieldSpecifications[i];
            if (spec.getType() == HasType.VALUE) {
                try {
                    NakedValue val = NakedObjectsContext.getObjectLoader().createValueInstance(spec);
                    val.parseTextEntry(entryText[i]);
                    entries[i] = val;
                } catch (InvalidEntryException e) {
                    errors[i] = e.getMessage();
                } catch (TextEntryParseException e) {
                    errors[i] = e.getMessage();
                }
            } else if (spec.getType() == HasType.OBJECT && entryText[i] != null) {
                if (entryText[i].equals("null")) {
                    entries[i] = null;
                } else {
                    entries[i] = context.getMappedObject(entryText[i]);
                }
            }
        }
        return entries;
    }

    public String getId() {
        return "" + id;
    }

    public boolean[] getOptional() {
        boolean[] array = new boolean[noOfEntriesInThisStep()];
        copyForThisStep(optional, array);
        return array;
    }

    public int[] getNoLines() {
        int[] array = new int[noOfEntriesInThisStep()];
        copyForThisStep(noLines, array);
        return array;
    }

    public boolean[] getWraps() {
        boolean[] array = new boolean[noOfEntriesInThisStep()];
        copyForThisStep(wraps, array);
        return array;
    }

    public int[] getMaxLength() {
        int[] array = new int[noOfEntriesInThisStep()];
        copyForThisStep(maxLength, array);
        return array;
    }

    public int[] getTypicalLength() {
        int[] array = new int[noOfEntriesInThisStep()];
        copyForThisStep(typicalLength, array);
        return array;
    }

    public boolean[] getReadOnly() {
        boolean[] array = new boolean[noOfEntriesInThisStep()];
        copyForThisStep(readOnly, array);
        return array;
    }

    public int getStep() {
        return step;
    }

    public NakedObject getTarget(Context context) {
        return context.getMappedObject(targetId);
    }

    public String[] getTrail() {
        String[] trail = new String[boundaries.length - 1];
        for (int i = 0; i < trail.length; i++) {
            trail[i] = "step " + i;
        }
        return trail;
    }

    public NakedObjectSpecification[] getTypes() {
        NakedObjectSpecification[] array = new NakedObjectSpecification[noOfEntriesInThisStep()];
        copyForThisStep(fieldSpecifications, array);
        return array;
    }

    public boolean isEditing() {
        return false;
    }

    public void nextStep() {
        step++;
    }

    private int noOfEntriesInThisStep() {
        return boundaries[step + 1] - boundaries[step];
    }

    public int numberOfSteps() {
        return boundaries.length - 1;
    }

    public void previousStep() {
        step--;
    }

    public void setFromFields(Request request, Context context) {
        int fldNo = 0;
        for (int i = boundaries[step]; i < boundaries[step + 1]; i++) {
            String textEntry = request.getFieldEntry(fldNo++);
            if (readOnly[i]) {
                continue;
            }
            NakedObjectSpecification spec = fieldSpecifications[i];
            // deal with check boxes specially: expect 'true' if checked and no entry if not checked, hence
            // need to set as 'false'
            if (spec.isOfType(NakedObjectsContext.getReflector().loadSpecification(boolean.class))
                    || spec.isOfType(NakedObjectsContext.getReflector().loadSpecification(Boolean.class))) {
                if (textEntry == null || !textEntry.equals("true")) {
                    textEntry = "false";
                }
            }
            entryText[i] = textEntry;
            try {
                errors[i] = null;
                setFromField(context, i, spec, textEntry);
                if (!optional[i] && (textEntry == null || textEntry.equals(""))) {
                    errors[i] = "Field required";
                }
            } catch (InvalidEntryException e) {
                errors[i] = e.getMessage();
            } catch (TextEntryParseException e) {
                errors[i] = e.getMessage();
            }
        }
    }

    private void setFromField(Context context, int i, NakedObjectSpecification type, String textEntry) {
        if (type.getType() == HasType.VALUE) {
            if (textEntry == null) {
                return;
            } else {
                // REVIEW this block uses the existing adapter as it contains the regex needed. This needs to
                // be reviewed in line with Dan's proposed changes to the reflector.
                NakedValue nakedValue = (NakedValue) initialState[i];
                if (nakedValue == null) {
                    nakedValue = NakedObjectsContext.getObjectLoader().createValueInstance(type);
                }
                String existingEntry = nakedValue.asEncodedString();
                try {
                    nakedValue.parseTextEntry(textEntry);
                } finally {
                    nakedValue.restoreFromEncodedString(existingEntry);
                }
            }
        }
    }

    public abstract void checkForValidity(Context context);

}

// Copyright (c) Naked Objects Group Ltd.
